c     Copyright (C)
c     2008-2013 Universit Paul Sabatier
c     2008-2013 Centre National de la Recherche Scientifique
c     2008-2013 Ecole des Mines d'Albi
c     2008-2013 Universit de Bordeaux 1
c     2008-2013 Fondation STAE
c     2014-2018 Meso-Star
c     
c     This file must be used under the terms of the CeCILL license.
c     This source file is licensed as described in the file COPYING, which
c     you should have received as part of this distribution.  The terms
c     are also available at
c     http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
c     
      subroutine line_truncation_sorting(ltr,dorg,tr_dist,ngamma,
     &     nu_min,nu_max,nuc,gamma,usl,line_index,
     &     T,sl_choice,slwr,slas,
     &     truncate)
      implicit none
      include 'max.inc'
c
c     Purpose: to sort lines because of line truncation; a given line (with central wavenumber nuc)
c     is considered relative to a wavenumber interval [numin,nu_max]
c     
c     Input:
c       + ltr: true if line profiles should be truncated
c       + dorg: truncate (1) depending on a fixed distance or (2) depending on a number of line width
c       + tr_dist: fixed truncation distance (inv. cm)
c       + ngamma: multiple of line width
c       + nu_min: lower limit of the considered wavenumber interval (inv. cm)
c       + nu_max: higher limit of the considered wavenumber interval (inv. cm)
c       + nuc: central wavenumber of the considered line (inv. cm)
c       + gamma: width of the considered line (inv. cm)
c       + usl: true if sub-lorentzian profile should be used
c       + line_index: index of the transition, as provided in the LBL database
c       + T: temperature (K)
c       + sl_choice: choice of sub-lorentzian profile
c         - 0: use whatever is relevant for current wavenumber, within validity ranges
c         - 1: use Pollack results @ 1.18m and 2.7m
c         - 2: use Perrin @ Hartmann @ 4.3m
c         - 3: use Tonkov @ 2.3m
c       + slwr: option for using CO2 SL profiles over the whole IR range
c       + slas: option to select using asymetric profile
c     
c     Output:
c       + truncate: true if the considered line should be truncated
c
c     NOTES: how truncation works now.
c     This routine is called by "constantk". Lines that should be truncated are added to
c     the constant lines over [nu_min,nu_max]. Lines that should not be truncated are then examined, and
c     those that do not vary a lot are also added to the constant lines over [nu_min,nu_max]. Finally, only
c     lines that should not be truncated and that vary over [nu_min,nu_max] are considered as non constant,
c     and their contributions are computed for every wavenumber within [nu_min,nu_max].
c     Additionnaly, the sum of constant lines contributions can be taken into account or not (in kspectrum).
c
c     Update 2015/05/07: I just realized that when using a sub-lorentzian profile (khi profile correction),
c     which may be the case for CO2 transitions, those corrected profiles should not be truncated, even if
c     truncation should be applied for other molecules.
c
c     I/O
      logical ltr
      integer dorg
      double precision tr_dist(1:Nmol_max)
      integer ngamma
      double precision nu_min,nu_max
      double precision nuc
      double precision gamma
      logical usl
      integer line_index
      double precision T
      integer sl_choice
      logical slwr,slas
      logical truncate
c     temp
      double precision khi_min,khi_max
c     label
      integer strlen
      character*(Nchar_mx) label
      label='subroutine line_truncation_sorting'

      truncate=.false.
      if (ltr) then
         if ((usl).and.(line_index.eq.2)) then
c     Special case of CO2: if a sub-lorentzian profile should be used, then DO NOT
c     truncate the line profile !
c     Debug - 14/12/2016
            
c     truncate=.false.
            call khi_co2(nu_min,nuc,T,sl_choice,slwr,slas,khi_min)
            call khi_co2(nu_max,nuc,T,sl_choice,slwr,slas,khi_max)
c     If a sub-lorentzian profile is used for line at 'nuc', then do not truncate
            if ((khi_min.lt.1.0D+0).or.(khi_max.lt.1.0D+0)) then
               truncate=.false.
c     Otherwise, truncate according to standard criteria (as for any other species)
            else
               if (dorg.eq.1) then
                  if ((nuc.lt.nu_min-tr_dist(2)).or.
     &                 (nuc.gt.nu_max+tr_dist(2))) then
                     truncate=.true.
                  endif
               else if (dorg.eq.2) then
                  if ((nuc.lt.nu_min-ngamma*gamma).or.
     &                 (nuc.gt.nu_max+ngamma*gamma)) then
                     truncate=.true.
                  endif
               else
                  call error(label)
                  write(*,*) 'dorg=',dorg
                  write(*,*) 'allowed values: 1, 2'
                  stop
               endif
            endif ! khi<1
c     Debug - 14/12/2016
c
         else
c     In all other cases (other species or no sub-lorentzian correction, even for CO2)
c     apply ordinary truncation scheme
            if (dorg.eq.1) then
               if ((nuc.lt.nu_min-tr_dist(line_index)).or.
     &              (nuc.gt.nu_max+tr_dist(line_index))) then
                  truncate=.true.
               endif
            else if (dorg.eq.2) then
               if ((nuc.lt.nu_min-ngamma*gamma).or.
     &              (nuc.gt.nu_max+ngamma*gamma)) then
                  truncate=.true.
               endif
            else
               call error(label)
               write(*,*) 'dorg=',dorg
               write(*,*) 'allowed values: 1, 2'
               stop
            endif
c
         endif ! (usl).and.(line_index=2)
c
      endif ! ltr

      return
      end


      
      subroutine line_truncation(ltr,dorg,tr_dist,ngamma,
     &     nu,nuc,gamma,usl,line_index,truncate)
      implicit none
      include 'max.inc'
c
c     Purpose: to perform all tests relative to line truncation for a given line
c     
c     Input:
c       + ltr: true if line profiles should be truncated
c       + dorg: truncate (1) depending on a fixed distance or (2) depending on a number of line width
c       + tr_dist: fixed truncation distance (inv. cm)
c       + ngamma: multiple of line width
c       + nu: wavenumber [cm]
c       + nuc: central wavenumber of the considered line (inv. cm)
c       + gamma: width of the considered line (inv. cm)
c       + usl: true if sub-lorentzian profile should be used
c       + line_index: index of the transition, as provided in the LBL database
c     
c     Output:
c       + truncate: true if the considered line should be truncated
c
c     I/O
      logical ltr
      integer dorg
      double precision tr_dist(1:Nmol_max)
      integer ngamma
      double precision nu
      double precision nuc
      double precision gamma
      logical usl
      integer line_index
      logical truncate
c     temp
c     label
      integer strlen
      character*(Nchar_mx) label
      label='subroutine line_truncation'

      truncate=.false.
      if (ltr) then
c         if ((usl).and.(line_index.eq.2)) then
cc     Special case of CO2: if a sub-lorentzian profile should be used, then DO NOT
cc     truncate the line profile !
c            truncate=.false.
c         else
            if (dorg.eq.1) then
               if (dabs(nu-nuc).ge.tr_dist(line_index)) then
                  truncate=.true.
               endif
            else if (dorg.eq.2) then
               if (dabs(nu-nuc).ge.ngamma*gamma) then
                  truncate=.true.
               endif
            else
               call error(label)
               write(*,*) 'dorg=',dorg
               write(*,*) 'allowed values: 1, 2'
               stop
            endif               ! dorg
c         endif                  ! (usl).and.(line_index=2)
      endif                     ! ltr

      return
      end
